# signature { license=0, seed=457995617, key="5f028b69069542dee51b5c70724c103e" }

local cgi = require('cgi')
local lfs = require('lfs')

local runningApps={}
local maxUploadSize = 400*1024








local cwd = package.path:gsub( '([?].*)', '' ):sub(1,-6)

local Z = package.config:sub(1,1)



local function map( ... )
	local f = table.pack( cwd, ... )
	return table.concat( f, Z ):gsub( '/', Z )
end

local attr = lfs.attributes

local function exists( ... ) return attr( map( ... ), 'size' ) end
local function isdir( ... ) 
	local dir = map( ... )
	return attr( dir, 'mode' ) == 'directory'
end

local lfsdir, lower
if Z=='/' then
	lfsdir = function( pattern )
		local path, mask = pattern:match( '^(.-/)([^/]+)$' )

		mask = mask and '^'..mask:gsub('[.]','%%.'):gsub('*','.-')..'$' or '.-'
		
		local dir, context = lfs.dir( path )
		local f = function( )
			while true do
				local name = dir( context )
				if not name then return nil end
				if name:find( mask ) then return name end
			end
		end
		return f
	end
	lower = function(s) return s end
else
	lfsdir = lfs.dir
	lower = string.lower
end

local function unescape(s)
	s = string.gsub( s, '+', ' ' )
	return string.gsub( s, "%%(%x%x)",
		function( hex )
			return string.char( tonumber( hex, 16 ) )
		end
	)
end





local i18nTexts = {}
do
	local code = loadfile( map( 'html/i18n.txt' ), 't', i18nTexts )
	if code then
		code()
		i18nTexts = i18nTexts.i18n
	end
end

function i18n( key, default ) return i18nTexts[ key ] or default or key end




permText =
{
	OS_MEMORY   = i18n("use up to %d bytes of main memory");
	OS_KILL     = i18n("stop other applications");
	LFS_MKDIR   = i18n("create directories on disk");
	LFS_RMDIR   = i18n("delete directories on disk");
	LFS_TOUCH   = i18n("change a file's time stamp");
	LFS_SETATTR = i18n("change a file's attributes");
	INST_LIB    = i18n("install a library module");
	UPD_APPMGR  = i18n("update the application manager");
}

setmetatable( permText, { __index = function( _, p ) return string.format( i18n("use system operation '%s'"), p ) end } )




local function rm_rf( path )
	if attr( path, 'mode' ) ~= 'directory' then return end
	
	require 'permissions'
	for file in lfsdir( path..Z..'*' ) do
		if file ~= '.' and file ~= '..' then
			local f = path..Z..file
			if attr( f, 'mode' ) == 'directory' then
				rm_rf( f )
			else
				os.remove( f )
			end
		end
	end
	
	lfs.rmdir( path )
end



local function saveTo( data, path )
	os.remove( path )
	local f = io.open( path, 'w' )
	if f then
		f:write( data )
		f:close()
		return true
	end
end





local function findRunningApps()
	runningApps = {}
	for i=1,10 do
		local state,path = os.psinfo(i)
		if state and state > 0 and state < 4 then
			local name= path:match( '[/\\]([^/\\]+)[/\\]%1[-.]lua$' )
			if name then
				runningApps[ name ] = i
				runningApps[ i ] = name
			end
		end
	end
end






local function preInstall( posted )
	require 'permissions'
	local zip = require( 'zip' )

	local tmp = map( 'html/manager.tmp' )
	local mnf = map( 'html/manifest.tmp' )
	saveTo( posted, tmp )
	local zipF = zip.open( tmp )
	if not zipF then return nil, i18n("package has wrong file format!") end

	local m = zipF:extract( 'manifest' )
	if not m or not m.data       then return nil, i18n("package contains no manifest") end
	if not saveTo( m.data, mnf ) then return nil, i18n("can't write manifest") end
	if os.license( mnf ) == nil  then return nil, i18n("manifest has an invalid signature") end


	local manifest = {}
	local code = loadfile( mnf, 't', manifest )
	if not code then return nil, i18n("manifest is not a valid Lua script") end
	code()
	local name = manifest.name  
	if not name then return nil, i18n("manifest contains no valid name") end


	lfs.mkdir( map( name ) )
	
	if isdir( name ) then

		if not exists( name, 'html/icon.png' ) then
			local icon = zipF:extract( 'html/icon.png' )
			if icon and icon.data then
				lfs.mkdir( map( name, 'html' ) )
				saveTo( icon.data, map( name, 'html/icon.png' ) )
			end
		end
	end
	zipF:close()
	

	os.rename( tmp, map( name, name..'.pkg' ) )
	os.rename( mnf, map( name, 'manifest.pkg' ) )
	
	return name
end

local function unInstall( name )
	if name == 'lib' or name == 'html' then return end -- these are protected

	if exists( name, 'uninstall.lua') then
		local uninstaller = map( name, 'uninstall.lua' )
		if os.license( uninstaller ) then pcall( dofile, uninstaller ) end
	end
	rm_rf( map( name ) )
end 

local function dontInstall( name )
	require 'permissions'
	
	if name ~= 'lib' and name ~= 'html' 
	and not exists( name, manifest ) then

		os.remove( map( name, 'html/icon.png' ) )
		os.rmdir( map( name, 'html' ) )
	end
	os.remove( map( name, 'manifest.pkg' ) )
	os.remove( map( name, name..'.pkg' ) )
end

local function doInstall( name )
	require 'permissions'
	local zip = require( 'zip' )

	local pkg = map( name, name..'.pkg' )
	local zipF, err = zip.open( pkg )
	if not zipF then
		MsgBox = i18n("zip error")..':\\n'..err
		return
	end

	local function unpackFiles( path, T, target )

		for name,subT in pairs( T ) do
			if type( subT ) == 'table' then
				lfs.mkdir( target..name )
				unpackFiles( path..name..'/', subT, target..name..Z)
			end
		end

		for _,name in ipairs( T ) do
			local fname = path..name
			local m = zipF:extract( fname )
			if m and m.data then saveTo( m.data, target..name ) end
		end
	end

	local mfname = map( name, 'manifest.pkg' )
	local manifest = {}
	local code = loadfile( mfname, 't', manifest )
	if code then
		code()
		local T = manifest.files
		unpackFiles( '', T, map( name, '' ) )
	end
	zipF:close()
	os.remove( pkg )
	os.remove( mfname )
	os.remove( map( name, name..'-lua' ) ) -- delete locked app


	if exists( name, 'install.lua') then
		local installer = map( name, 'install.lua' )
		if os.license( installer ) then pcall( dofile, installer ) end
	end
end

local function getManifest( p )
	local M = {}
	local code = loadfile( p, 't', M )
	if code then
		code()
		return M
	end
end

local function getAppDetails( name )
	if name=='.' or name=='..' or not isdir( name ) then return end
	
	local app = {
		name   = name ,
		icon   = exists( name, 'html/icon.png' ) ,
		config = exists( name, 'html/config.cgi' ) ,
		M      = getManifest( map( name, 'manifest' ) ) ,
		N      = getManifest( map( name, 'manifest.pkg' ) )
	}
	app.license = (app.N or app.M or {license=3500}).license or 0
		
	if exists( name, name..'.lua' ) then
		app.canRun = true
	elseif exists( name, name..'-lua' ) then
		app.isLocked = true
		app.canRun = true
	end

	if exists( name, 'html', name..'.cgi' )  then app.home = name..'.cgi' end
	if runningApps[ name ] then app.canStop = true	end
	
	if app.N then return app end
	if not app.M then return end
	if name == 'lib' or name == 'html' then return end

	return app
end

local function paragraph( title, formatstring, ... )
	local t = select( 1, ... )
	if not t or t == '' then return '' end
	t = string.format( formatstring, ...)
	cgi.printf( '<h5>%s</h5>%s<br />', title, t or '')
end

local function listPermissions( P )
	local ph=false;
	for k,v in pairs(P) do
		if not ph then
			cgi.print( '<h5>', i18n("Required Permissons"), '</h5><br />', i18n("This application requires your permission to"), '<ul>')
			ph = true
		end
		cgi.print( '<li>' )
		cgi.printf( permText[ k ], v )
		cgi.print( '</li/>' )
	end
	if ph then cgi.print( '</ul>' ) end
end

local function listRequirements( R )
	local function get( name )
		local loaded = package.loaded
		package.loaded = {}
		local ok, L = pcall( require, name )
		package.loaded = loaded
		if ok then return L end
	end

	if R.licenses then
		local ph = false
		for k,v in pairs( R.licenses ) do
			local n = os.license( tonumber( k ) ) 
			local V = tonumber( v ) or 1
			if n < V then
				if not ph then
					cgi.print('<h5>', i18n("Required Licenses"), '</h5><br />', i18n("This application requires to install"),"<ul>")
					ph = true;
				end
				cgi.printf( '<li>%d %s %d</li/>', V-n, i18n("instance(s) of License"), k )
			end
		end
		if ph then cgi.print( '</ul>' ) end
	end

	if R.libraries then
		local ph = false
		for k,v in pairs( R.libraries ) do
			local V = v
			local lib = get( k )
			local versionOK
			if not lib then 
				versionOK = false
				V = ''
			else
				local imajor,iminor = (lib._VERSION or 'version 0.0' ):match( '^%w+%s+(%d+)[.]?(%d*)' )
				versionOK = type( v ) ~= 'string'
				imajor = tonumber( imajor or 0 )
				iminor = tonumber( iminor or 0 )
				if versionOK then 
					V = '0.0'
				else
					local wmajor,wminor = v:match( '(%d+)[.]?(%d*)' )
					wmajor = tonumber( wmajor or 0 )
					wminor = tonumber( wminor or 0 )
					versionOK =  imajor > wmajor or
					imajor == wmajor and
					iminor >= wminor 
				end
			end
			if not versionOK then
				if not ph then
					cgi.print( '<h5>', i18n("Required Libraries"), '</h5><br />', i18n("This application requires to install"), '<ul>')
					ph = true;
				end
				cgi.printf( '<li>%s %s</li/>', k, V)
			end
			lib = nil

			collectgarbage()
		end
		if ph then cgi.print( '</ul>' ) end
	end
end

local function geti18n()
	local set = {}

	local file = io.open( map( 'html', 'manager.cgi' ) )
	local text = file:read( '*a' )
	file:close()

	for _, key in text:gmatch( 'i18n%s*[(]%s*([\"\'])(.-)%1' ) do 
		set[ key ] = key
	end

	cgi.header( 'Content-Disposition', 'attachment; filename=i18n.txt' );

	for row in io.lines( map( 'html', 'i18n.txt' ) ) do
		local line = row:gsub( '\r', '' )
		local key = line:match( '%["([^"]*)"%]' )
		if key then 
			if set[key] then
				set[key] = nil
			else
				line = line .. "-\45 unused !"
			end
		else
			if line:find( '^}' ) then -- end of array, add new entries
				local sorted = {}
				for key in pairs( set ) do table.insert( sorted, key ) end

				cgi.printf( "-\45\r\n-\45 new texts\r\n-\45\r\n" )
				table.sort( sorted )
				for _, key in ipairs( sorted ) do
					cgi.printf( '["%s"]=\r\n[[%s]];\r\n\r\n' , key, key:gsub( '\\n', '\r\n' ) )
				end
			end
		end
		cgi.printf( '%s\r\n', line )
	end
	return true
end




local authorised = false
 
local function auth()
	local port = require( 'xmlpro.core' ).port
	local socket = require('socket')
	local text = '<get><mb mb=\"9002\"><k k=\"2\">[1-2]</k></mb></get>'
	local request = ( 'POST /dbms.xml HTTP/1.1\r\nAuthorization: %s\r\nContent-Length: %d\r\n\r\n%s' ):format( CGI.AUTHORIZATION or '', #text, text )
	local s = socket.connect( '127.0.0.1', port )
	s:send( request )

	local response = ''
	local status = s:receive():match(" (%d%d%d)")
	if "200" == status then
		local len=0
		repeat 
			local line = s:receive()  -- receive header lines
			len = line:lower():match( "^content%-length:%s*(%d+)") or len
		until line==""

		response = s:receive( tonumber(len) or 0 )
	end
	s:close()
	return response, status
end

if not authorised then

	local x = auth()
	if string.find( x, 'i=' ) then
		authorised = true
	else
		require( 'socket' ).sleep( 3 )
	end
end

if not authorised then
	cgi.status = "401 Unauthorized"
	cgi.header( 'WWW-Authenticate', 'Basic realm="NEC Application Manager"' )
	cgi.print( "Unauthorised!" )
	return
end

findRunningApps()
local cmd,value = '', ''
if CGI.REQUEST_METHOD == 'POST' then
	local name, posted, err
	local length = tonumber( CGI.CONTENT_LENGTH ) or 0

	if length > maxUploadSize then 
		MsgBox = i18n("Package is too large")..'\\n'


		while length>8192 do
			cgi.receive( 8192 )
			length = length - 8192
		end
		cgi.receive( length )
		length = 0
	end

	if length > 0 then
		posted = cgi.receive( length )
		local enctype, boundary = string.match( (CGI.CONTENT_TYPE or ''), '%s*([^;]+);%s*boundary=(.*)$' )
		if enctype == 'multipart/form-data' then
			boundary = string.gsub( boundary, '[][().%+-*?^$]', '%%%1' )
			posted = string.match( posted, boundary..'.-\r\n\r\n(.-)\r\n%-%-'..boundary )
			name, err = preInstall( posted )
			if name then 
				value = name
			else
				MsgBox = i18n("Package upload failed")..'\\n'..err
			end
		else
			posted = unescape( posted )
			cmd, value = string.match( posted, 'cmd=([^=]+)=(.+)' )
		end
	end
elseif CGI.QUERY_STRING > '' then
	cmd, value = string.match( CGI.QUERY_STRING, '&?([^=]+)=([^&]+)' )
end

local has3500 = os.license( 3500 ) > 0




local commands = {
	run = function( value )
		if not runningApps[ value ] then 
			local name 
			if exists( value, value .. '.lua' ) then
				name = '.lua'
			elseif exists( value, value..'-lua' ) then
				name = '-lua'
			else 
				return
			end
			name = map( value, value..name )

			local i = os.execute( name )
			runningApps[ value ] = i
			runningApps[ i ] = value
		end
		return true
	end ,

	stop = function( value )
		require 'permissions'
		local i = runningApps[ value ]
		if i then
			os.kill( i )
			findRunningApps()
			runningApps[ i ]     = nil -- it will be stopped soon
			runningApps[ value ] = nil -- it will be stopped soon
		end
	end ,

	remove = unInstall ,

	cancel = dontInstall ,
	install = doInstall ,
	lock   = function( value ) os.rename( map(  value, value..'.lua' ) , map( value, value..'-lua' ) ) return true end ,
	unlock = function( value ) os.rename( map(  value, value..'-lua' ) , map( value, value..'.lua' ) ) return true end ,
	no3500 = function() has3500 = false print( "Developer License 3500 ignored" ) end ,
	geti18n = geti18n ,
}

local f = commands[ cmd ]
if f and  f( value ) then 
	cgi.header( 'Content-Type', 'text/plain; charset="UTF-8"' )
	return
end


local apps = {}
for Name in lfsdir( map('*') ) do
	local name = lower( Name )
	local app = getAppDetails( name )
	if app then
		table.insert( apps, app )
		apps[ app.name ] = app
	end
end

local selected = { }
if #apps > 0 then
	if value == '' or not apps[ value ] then value = apps[1].name end
	selected[ value ] = 'sel'
end




cgi.header( 'Content-Type', 'text/html; charset="UTF-8"' )
cgi.header( 'X-UA-Compatible', 'IE=edge' ) -- override IE's "show intranet sites in compatibility mode"
cgi.printf( [[
<!DOCTYPE HTML>
<html>
<head>
<meta charset="UTF-8" />
<title>%s</title>
<link rel="SHORTCUT ICON" href="/html/apps/favicon.ico">
<link rel="stylesheet" type="text/css" href="common.css" >
<link rel="stylesheet" type="text/css" href="manager.css" >
<script>var MsgBox="%s";
var confirm1="%s";
var confirm2="%s";
]], 
	i18n("NEC PBX Application Manager"),
	MsgBox or '',
	i18n("Do you really want to remove this application\nand all its files and directories ?"):gsub('\n', '\\n'),
	i18n("Are you sure ?")
)

cgi.print( [[var AppNames=[ ]] )
local comma = ''
for _,app in ipairs( apps ) do
	cgi.printf( '%s"%s"', comma, app.name )
	comma = ','
end
cgi.print( ' ];</script>' )


cgi.printf( [[
<script src="manager.js" type="text/javascript" ></script>
</head>
<body>
	<h1>%s <small>%s</small></h1>
	<h2>
	  <form name="upload" action="manager.cgi" method="post" enctype="multipart/form-data">
        <a class="button"><div class="upload">
          <input type="file" id="uploadfile" name="uploadfile"/>%s</div>
        </a>
	  </form>
      <a class="button hide" id="refresh" href="javascript:doPoll();"><img src="refresh.png" />%s</a>
      %s</h2>
	<div class="apps">
		<form name="post" action="manager.cgi" method="post"><input type="hidden" name="cmd" /></form>
		<ul class="apps" id="applist">
]],i18n("Application Manager"), "1.5.1", i18n("Install"), i18n("Refresh"), i18n("Installed Applications"))




for _,app in ipairs( apps ) do
	local M = app.N or app.M
	
	cgi.printf( '<li  style="background-image:url(/html/apps%s/icon.png);" onclick="details(\'%s\')" id="list-%s" class="%s"><h3>%s&nbsp;<small>%s</small></h3>%s</li>\r\n',
		app.icon and '/'..app.name or '',
		app.name,    -- onclick()
		app.name,    -- id
		selected[ app.name ] or '',
		M.description.name   or i18n("Applet"), -- H4 title
		M.version or '',
		M.description.summary or ''
	)
end

cgi.print([[</ul></div>
<div class="details" id="appdetails">]])




	for _,app in ipairs(apps) do
		local M = app.N or app.M
		local c = M.description
		local licenses 

		if app.license == 0 or has3500 then
			licenses = 1 -- always on
		else
			licenses = os.license( app.license )
			if licenses < 1 then
				app.canRun = false
				app.config = false
			end
		end

		cgi.printf( '<div id="detail-%s" class="%s" data-state="5"><h4 class="buttons">%s ',
			app.name,
			selected[ app.name ] or '',
			c.name   or i18n("Details") -- H4 title
		)

		if app.name ~= 'html' and app.name ~= 'lib' then

			if app.config then
				local cfgrun = app.M.RCPORT and ' state3' or ''
				cgi.printf([[<a class="button%s" id="%s-config" href="/html/apps/%s/config.cgi" target="_blank"><img src="config.png" />%s</a>]], cfgrun, app.name, app.name, i18n("Configure"))
			end

			if app.canRun then
				cgi.printf([[<a class="button state0" id="%s-run" href="javascript:doRun('%s')"><img src="play.png" />%s</a>]], app.name, app.name, i18n("Start"))
			end

			cgi.printf([[<a class="button state3 state4"  id="%s-stop"   href="javascript:doStop('%s')"><img src="stop.png" />%s</a>]], app.name, app.name, i18n("Stop"))
			if app.M then
				cgi.printf([[<a class="button state0"  id="%s-remove"   href="javascript:doRemove('%s')"><img src="trash.png" />%s</a>]], app.name, app.name, i18n("Remove"))

				if app.canRun and licenses > 0 then
					cgi.printf([[<a class="button%s" id="%s-unlock" href="javascript:doUnlock('%s')"><img src="unlock.png" />%s</a>]], app.isLocked and "" or " hide", app.name, app.name, i18n("Enable"))
					cgi.printf([[<a class="button%s" id="%s-lock"   href="javascript:doLock('%s')"><img src="lock.png" />%s</a>]],    app.isLocked and " hide" or "", app.name, app.name,i18n("Disable"))
				end
			end

			if app.M then -- installed
				for _,button in ipairs(c.buttons or {}) do
					if button.img and button.href then
						cgi.printf( '<a class="button %s" href="%s"', button.class or "", button.href )
						if button.target then
							cgi.printf( ' target="%s"', button.target)
						end
						if button.id then
							cgi.printf( ' id="%s-%s"', app.name, button.id)
						end
						cgi.printf( '><img src="%s" />%s</a>', button.img, i18n( button[1] or "" ) )
					end 
				end
			end
		end

		cgi.print( '</h4>' )

		paragraph( i18n("Summary:"), '<br />%s', c.summary)
		paragraph( i18n("Description:"), '<br />%s', c.detailed:gsub('<script>.-</script>', '') )
		paragraph( i18n("Home Page:"), '<a href="%s">%s</a>', c.homepage, c.homepage)
		paragraph( i18n("Maintained by:"), '%s', c.maintainer)
		paragraph( i18n("Update URL:"), '<a href="%s">%s</a>', c.updateURL, c.updateURL)
		paragraph( i18n("Required License Code:"), '%s', (app.license==0) and i18n("free") or tostring(app.license) )
		if licenses <= 0 then
			cgi.printf( '<em>%s<br />%s</em><br />',	i18n("This license is not installed."), i18n("It is required to start or configure the application"))
		end
		paragraph( i18n("Copyright:"), '%s', c.copyright)

		if app.N then
			cgi.printf( [[<h4 class="buttons hr">%s
<a class="button" id="%s-install" href="javascript:doInstall('%s')"><img src="ok.png" />%s</a>
<a class="button" id="%s-cancel" href="javascript:doCancel( '%s')"><img src="cancel.png" />%s </a></h4>]],
				i18n("Uploaded Package ready for Installation"),
				app.name,	app.name, i18n("Install"),
				app.name,	app.name, i18n("Cancel")
			)
			if app.N.permissions then listPermissions( app.N.permissions ) end
			if app.N.requirements then listRequirements( app.N.requirements ) end
			if app.N.version then paragraph( i18n("New Version:"), "%s", app.N.version) end
		end
		if app.M then
			paragraph( i18n("Installed Version:"), '%s', app.M.version or '')
		end
		cgi.print( '</div>\r\n' )
	end
	cgi.print('</div></body></html>')

















