# signature { license=0, seed=457995617, key="33af4663a9c86902f03a6db21c45e3f3" }
--




local zlib = require('zlib')
local shl, shr, bor, band = bit32.lshift, bit32.rshift, bit32.bor, bit32.band


local pack, unpack = string.pack, string.unpack




local function closed() return nil, "closed" end



local function closeArchive( archive )
	archive.file:close()
	archive.file = nil
	archive.dir = nil
	archive.write = closed
	archive.read = closed
end



local function noCompression( data, flush ) return data, flush=='finish', #data, #data end



local function readArchive( archive, length )
	return archive.file:read( length )
end





local function readZippedFile( T, length )
	local data = T.archive:read( length or T.cmpSize )
	if data then 
		return T.inflate( data )
	end
end



local function closeZippedFile( T )
	if T then
		T.close = closed
		T.read  = closed
	end
end



local openedZipMT =
{
	__index =
	{
		read  = readZippedFile ;
		close = closeZippedFile ;
	}
}




local function openZippedFile( archive, fileName )
	local T = archive.dir[fileName]
	if not T then return nil, "file not in ZIP" end
	
	local zipF = archive.file
    zipF:seek( 'set', T.localFileHeader )
	
	local nameLen, extraLen
	local s = zipF:read(30)
	if s and #s==30 and s:sub(1,4) == 'PK\3\4' then
		nameLen, extraLen = unpack( '<HH', s, 27 )
		zipF:read( nameLen + extraLen )
	

		T.inflate = T.compression == 8 and zlib.inflate(-15) or noCompression
		
		return setmetatable( T, openedZipMT )
	else
		return nil, "not a local file header"
	end
end





local function extractZippedFile( archive, name )
	local T, err = openZippedFile(  archive, name )
	
	if T then
		return { dir = T ; data = T:read( T.cmpSize ) ; }
	else
		return nil, err
	end
end              








local function writeArchive( archive, data )
	if data and #data>0 then
		archive.file:write( data )
		archive.wpos = archive.wpos + #data
	end
end



local function processZippedData( T, data, flush )
	if T and data then
		T.CRC = T.crc32( data )
		local compressed, _
		compressed, _, T.orgSize , T.cmpSize = T.deflate( data, flush )
		return compressed
	end
end



local function writeZippedFile( T, data, flush )
	if T and data then
		T.archive:write( processZippedData( T, data, flush ) )
	end
	return T
end




















local function writeLocalHeader( T )
	T.archive:write
	(
		pack(
			'<cHHHHHHLLHxxc' ,
			'PK\3\4\20\0' ,	-- local file header signature, version
			T.gpflag ,
			T.compression ,
			T.mTime ,
			T.mDate ,
			band( T.CRC, 0xFFFF ), shr( T.CRC, 16 ) , 
			T.cmpSize ,
			T.orgSize ,
			#T.fileName ,
			T.fileName
		)
	)
end




local function completeZippedFile( T, enbloc )
	if T then
		if not enbloc then
			T:write( '', 'finish' )
			if T.streamed then

				T.archive:write
				(
					pack
					(
						'<cHHLL' ,
						'PK\7\8' ,
						band( T.CRC, 0xFFFF), shr( T.CRC, 16) , 
						T.cmpSize ,
						T.orgSize 
					)
				)
			else

				local wpos = T.archive.wpos
				T.archive.file:seek( 'set', T.localFileHeader +14)
				T.archive:write
				(
					pack
					(
						'<HHLL' ,
						band( T.CRC, 0xFFFF), shr( T.CRC, 16) , 
						T.cmpSize ,
						T.orgSize 
					)
				)
				T.archive.file:seek( 'set', wpos )
				T.archive.wpos = wpos
			end
		end
		T.close = closed
		T.write = closed
	end
end



local createdZipMT =
{
	__index =
	{
		write = writeZippedFile ;
		close = completeZippedFile ;
	}
}



local function createDirEntry( archive, fileName, uncompressed, mtime )
	if not archive or not fileName then return end
	
	local ts = os.date( mtime and '!*t' or '*t', mtime )
	
	local T = 
	{ 
		archive = archive , 
		fileName = fileName , 
		gpflag = 2 ,
		mTime   = shl(ts.hour , 11) + shl( ts.min, 5) + ts.sec ,
		mDate   = shl(ts.year - 1980, 9) + shl( ts.month, 5) + ts.day ,
		localFileHeader = archive.wpos ,
		CRC = 0 ,
		cmpSize = 0 ,
		orgSize = 0 ,
		compression = uncompressed and 0 or 8 ,

		deflate = uncompressed and noCompression or zlib.deflate(9,-15) ,
		crc32 = zlib.crc32() , 
	}
	local dir =	archive.dir
	dir[ T.fileName ] = T
	dir[ #dir +1 ] = T

	return setmetatable( T, createdZipMT )
end












local function createZippedFile( archive, fileName, uncompressed, mtime )
	local T = createDirEntry( archive, fileName,  uncompressed, mtime )
	if not T then return end
	
	if not T.archive.file.seek then
		T.streamed = true
		T.gpflag = 10
	end
	
	writeLocalHeader( T )
	
	return setmetatable( T, createdZipMT )
end











local function storeZippedFile( archive, fileName, data, uncompressed, mtime )
	local T = createDirEntry( archive, fileName, uncompressed, mtime )
	if T then
		local zipped = processZippedData( T, data, 'finish' )
		writeLocalHeader( T )
		T.archive:write( zipped )
		T:close( true )
	end
end




















































local function completeArchive( archive )
	if not archive or not archive.file then return end
	
	local cDirOffset = archive.wpos
	local cDirSize = 0
	local files = 0
	

	for _,T in ipairs( archive.dir ) do
		local entry = 
			pack
			(
				'<cHHHHHHLLHxxxxxxxxxxxxLc' ,

				'PK\1\2\20\0\20\0' ,	
				T.gpflag ,
				T.compression, 
				T.mTime ,
				T.mDate ,
				band( T.CRC, 0xFFFF), shr( T.CRC, 16) , 
				T.cmpSize ,
				T.orgSize ,
				#T.fileName ,
				T.localFileHeader ,
				T.fileName
			)
		archive:write( entry )
		
		cDirSize = cDirSize + #entry
		files = files +1
	end
	

	archive:write
	(
		pack
		(
			'<cHHLLxx', 
			'PK\5\6\0\0\0\0',
			files,
			files,
			cDirSize,
			cDirOffset
		)
	)
	
	return closeArchive( archive )
end



local openedArchiveMT = 
{
	__index =
	{
		read    = readArchive ;
		close   = closeArchive ;
		
		extract = extractZippedFile ; 
		open    = openZippedFile ;
	} ;
	__gc = function(self) pcall( closeArchive, self) end ;
}










local function openArchive( zipF )
	if type (zipF) == 'string' then zipF = io.open( zipF, 'rb') end
	if not zipF then return nil, "ZIP file not found" end
	

	zipF:seek( 'end', -22)
	local s = zipF:read( 22 )
	if not s or #s~=22 or s:sub(1,4) ~= 'PK\5\6' then return nil, "not a ZIP file" end
	
	local nEntries, cDirOffset = unpack( '<HxxxxL', s, 11 )
	
	zipF:seek( 'set', cDirOffset ) 
	
	local dir = {}
	local archive =  { file = zipF, dir = dir }
	

	local MT = 
	{ 
		__index =
		{
			archive = archive ;
			
			extract = extractZippedFile ;
			open    = openZippedFile ;
			read    = readZippedFile ;
			close   = closeZippedFile ;
		}
	}
	

	for i=1,nEntries do
		local nameLen, extraLen, commentLen;
		local T = setmetatable( {}, MT )
		
		local s = zipF:read(46)
		
		if s and #s==46 and s:sub(1,4) == 'PK\1\2' then
			T.gpflag ,
			T.compression ,
			T.mTime ,
			T.mDate ,
			T.CRC , 
			T.cmpSize ,
			T.orgSize ,
			nameLen ,
			extraLen ,
			commentLen ,
			T.localFileHeader =
			unpack( '<HHHHLLLHHHxxxxxxxxL', s, 9 )
			

			T.CRC = bor( T.CRC, 0 )
			T.archive = archive
		else
			return nil, "invalid central directory entry" 
		end

		T.fileName = zipF:read( nameLen )
		
		dir[ T.fileName ] = T
		dir[ i ] = T
		
		zipF:read( extraLen + commentLen )
	end

	return setmetatable( archive, openedArchiveMT )
end




local createdArchiveMT = 
{
	__index =
	{
		write  = writeArchive ;
		close  = completeArchive ;
		
		store  = storeZippedFile ; 
		create = createZippedFile ;
	} ;
	__gc = function(self) pcall( completeArchive, self) end ;
}





















local function createArchive( zipF )
	if type( zipF ) == 'function' then 
		zipF = { write = zipF, close = function() end } 
	elseif type( zipF ) == 'string' then 
		zipF = io.open( zipF, 'wb' )
	end
	
	if not zipF then return nil, "can't create ZIP file" end

	local archive = 
	{
		file = zipF ;
		dir = {} ;
		wpos = 0 ;
	}
	
	return setmetatable( archive, createdArchiveMT )
end




local zip = 
{
	_VERSION = 'zip 0.6.3' ;
	_COPYRIGHT = 'NEC 2017' ;
	_DESCRIPTION = 'Module to create and write or open and read ZIP archives' ;
	
	open   = openArchive ;
	create = createArchive ;
}

return zip
